/**
 * Bluetooth Asia 2019 - Bluetooth Mesh Provisioning and Interoperability Session
 *
 * Bluetooth Mesh Provisioning Sample
 *
 **/

#include <stdio.h>

#include <gpio.h>
#include <ctype.h>

#include <misc/printk.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/mesh.h>

#include <device.h>
#include <init.h>
#include <soc.h>


#define MODE_MASK    BIT_MASK(16)

/* Onboard LED Row 1 */
#define LED_ROW1_GPIO_PIN   13
#define LED_ROW1_GPIO_PORT  DT_GPIO_P0_DEV_NAME

/* Onboard LED Row 2 */
#define LED_ROW2_GPIO_PIN   14
#define LED_ROW2_GPIO_PORT  DT_GPIO_P0_DEV_NAME

/* Onboard LED Row 3 */
#define LED_ROW3_GPIO_PIN   15
#define LED_ROW3_GPIO_PORT  DT_GPIO_P0_DEV_NAME

/* Onboard LED Column 1 */
#define LED_COL1_GPIO_PIN   4
#define LED_COL1_GPIO_PORT  DT_GPIO_P0_DEV_NAME

/* Onboard LED Column 2 */
#define LED_COL2_GPIO_PIN   5
#define LED_COL2_GPIO_PORT  DT_GPIO_P0_DEV_NAME

/* Onboard LED Column 3 */
#define LED_COL3_GPIO_PIN   6
#define LED_COL3_GPIO_PORT  DT_GPIO_P0_DEV_NAME

/* Onboard LED Column 4 */
#define LED_COL4_GPIO_PIN   7
#define LED_COL4_GPIO_PORT  DT_GPIO_P0_DEV_NAME

/* Onboard LED Column 5 */
#define LED_COL5_GPIO_PIN   8
#define LED_COL5_GPIO_PORT  DT_GPIO_P0_DEV_NAME

/* Onboard LED Column 6 */
#define LED_COL6_GPIO_PIN   9
#define LED_COL6_GPIO_PORT  DT_GPIO_P0_DEV_NAME

/* Onboard LED Column 7 */
#define LED_COL7_GPIO_PIN   10
#define LED_COL7_GPIO_PORT  DT_GPIO_P0_DEV_NAME

/* Onboard LED Column 8 */
#define LED_COL8_GPIO_PIN   11
#define LED_COL8_GPIO_PORT  DT_GPIO_P0_DEV_NAME

/* Onboard LED Column 9 */
#define LED_COL9_GPIO_PIN   12
#define LED_COL9_GPIO_PORT  DT_GPIO_P0_DEV_NAME

#define BT_MESH_MODEL_OP_GEN_ONOFF_GET		BT_MESH_MODEL_OP_2(0x82, 0x01)
#define BT_MESH_MODEL_OP_GEN_ONOFF_SET		BT_MESH_MODEL_OP_2(0x82, 0x02)
#define BT_MESH_MODEL_OP_GEN_ONOFF_SET_UNACK	BT_MESH_MODEL_OP_2(0x82, 0x03)
#define BT_MESH_MODEL_OP_GEN_ONOFF_STATUS	BT_MESH_MODEL_OP_2(0x82, 0x04)

static struct bt_mesh_cfg_srv cfg_srv = {
	.relay = BT_MESH_RELAY_DISABLED,
	.beacon = BT_MESH_BEACON_ENABLED,
#if defined(CONFIG_BT_MESH_FRIEND)
	.frnd = BT_MESH_FRIEND_ENABLED,
#else
	.frnd = BT_MESH_FRIEND_NOT_SUPPORTED,
#endif
#if defined(CONFIG_BT_MESH_GATT_PROXY)
	.gatt_proxy = BT_MESH_GATT_PROXY_ENABLED,
#else
	.gatt_proxy = BT_MESH_GATT_PROXY_NOT_SUPPORTED,
#endif
	.default_ttl = 7,

	/* 3 transmissions with 20ms interval */
	.net_transmit = BT_MESH_TRANSMIT(2, 20),
	.relay_retransmit = BT_MESH_TRANSMIT(2, 20),
};

static u8_t genericOnOffstatus;

static struct device *gpio;

void allleds_on(void)
{
	gpio_pin_write(gpio, LED_ROW1_GPIO_PIN, 1);
	gpio_pin_write(gpio, LED_ROW2_GPIO_PIN, 1);
	gpio_pin_write(gpio, LED_ROW3_GPIO_PIN, 1);
	gpio_pin_write(gpio, LED_COL1_GPIO_PIN, 0);
	gpio_pin_write(gpio, LED_COL2_GPIO_PIN, 0);
	gpio_pin_write(gpio, LED_COL3_GPIO_PIN, 0);
	gpio_pin_write(gpio, LED_COL4_GPIO_PIN, 0);
	gpio_pin_write(gpio, LED_COL5_GPIO_PIN, 0);
	gpio_pin_write(gpio, LED_COL6_GPIO_PIN, 0);
	gpio_pin_write(gpio, LED_COL7_GPIO_PIN, 0);
	gpio_pin_write(gpio, LED_COL8_GPIO_PIN, 0);
	gpio_pin_write(gpio, LED_COL9_GPIO_PIN, 0);
}

void allleds_off(void)
{
	gpio_pin_write(gpio, LED_ROW1_GPIO_PIN, 0);
	gpio_pin_write(gpio, LED_ROW2_GPIO_PIN, 0);
	gpio_pin_write(gpio, LED_ROW3_GPIO_PIN, 0);
	gpio_pin_write(gpio, LED_COL1_GPIO_PIN, 1);
	gpio_pin_write(gpio, LED_COL2_GPIO_PIN, 1);
	gpio_pin_write(gpio, LED_COL3_GPIO_PIN, 1);
	gpio_pin_write(gpio, LED_COL4_GPIO_PIN, 1);
	gpio_pin_write(gpio, LED_COL5_GPIO_PIN, 1);
	gpio_pin_write(gpio, LED_COL6_GPIO_PIN, 1);
	gpio_pin_write(gpio, LED_COL7_GPIO_PIN, 1);
	gpio_pin_write(gpio, LED_COL8_GPIO_PIN, 1);
	gpio_pin_write(gpio, LED_COL9_GPIO_PIN, 1);
}

static void gen_onoff_get(struct bt_mesh_model *model,
			  struct bt_mesh_msg_ctx *ctx,
			  struct net_buf_simple *buf)
{
	printk("gen_onoff_get \n");
	NET_BUF_SIMPLE_DEFINE(msg, 2 + 1 + 4);

	bt_mesh_model_msg_init(&msg, BT_MESH_MODEL_OP_GEN_ONOFF_STATUS);
	net_buf_simple_add_u8(&msg, genericOnOffstatus);

	if (bt_mesh_model_send(model, ctx, &msg, NULL, NULL)) {
		printk("Unable to send On Off Status response");
	}
}


static void gen_onoff_set_unack(struct bt_mesh_model *model,
				struct bt_mesh_msg_ctx *ctx,
				struct net_buf_simple *buf)
{
	genericOnOffstatus = buf->data[0];

	printk("gen_onoff_set_unack, len: 0x%04X value: 0x%02X \n", buf->len, genericOnOffstatus);

	if(genericOnOffstatus){
		allleds_on();
	}
	else{
		allleds_off();
	}
}

static void gen_onoff_set(struct bt_mesh_model *model,
			  struct bt_mesh_msg_ctx *ctx,
			  struct net_buf_simple *buf)
{
	printk("gen_onoff_set\n");
	gen_onoff_set_unack(model, ctx, buf);
	gen_onoff_get(model, ctx, buf);
}

static const struct bt_mesh_model_op gen_onoff_op[] = {
	{ BT_MESH_MODEL_OP_2(0x82, 0x01), 0, gen_onoff_get },
	{ BT_MESH_MODEL_OP_2(0x82, 0x02), 2, gen_onoff_set },
	{ BT_MESH_MODEL_OP_2(0x82, 0x03), 2, gen_onoff_set_unack },
	BT_MESH_MODEL_OP_END,
};


static struct bt_mesh_model root_models[] = {
	BT_MESH_MODEL_CFG_SRV(&cfg_srv),
	BT_MESH_MODEL(BT_MESH_MODEL_ID_GEN_ONOFF_SRV, gen_onoff_op,
		      NULL, NULL),
};

static struct bt_mesh_elem elements[] = {
	BT_MESH_ELEM(0, root_models, BT_MESH_MODEL_NONE),
};

static const struct bt_mesh_comp comp = {
	.cid = BT_COMP_ID_LF,
	.elem = elements,
	.elem_count = ARRAY_SIZE(elements),
};

static int output_number(bt_mesh_output_action_t action, uint32_t number)
{

	return 0;
}

static int output_string(const char *str)
{
	printk("OOB String: %s\n", str);

	return 0;
}

static void prov_complete(u16_t net_idx, u16_t addr)
{
}

static void prov_reset(void)
{
	bt_mesh_prov_enable(BT_MESH_PROV_ADV | BT_MESH_PROV_GATT);
}

static uint8_t dev_uuid[16] = { 0xdd, 0xdd };

static const struct bt_mesh_prov prov = {
	.uuid = dev_uuid,
	.output_size = 4,
	.output_actions = BT_MESH_NO_OUTPUT,
	.output_number = output_number,
	.output_string = output_string,
	.complete = prov_complete,
	.reset = prov_reset,
};

void board_init(void)
{
	gpio = device_get_binding(DT_GPIO_P0_DEV_NAME);
	gpio_pin_configure(gpio, LED_ROW1_GPIO_PIN, GPIO_DIR_OUT);
	gpio_pin_configure(gpio, LED_ROW2_GPIO_PIN, GPIO_DIR_OUT);
	gpio_pin_configure(gpio, LED_ROW3_GPIO_PIN, GPIO_DIR_OUT);
	gpio_pin_configure(gpio, LED_COL1_GPIO_PIN, GPIO_DIR_OUT);
	gpio_pin_configure(gpio, LED_COL2_GPIO_PIN, GPIO_DIR_OUT);
	gpio_pin_configure(gpio, LED_COL3_GPIO_PIN, GPIO_DIR_OUT);
	gpio_pin_configure(gpio, LED_COL4_GPIO_PIN, GPIO_DIR_OUT);
	gpio_pin_configure(gpio, LED_COL5_GPIO_PIN, GPIO_DIR_OUT);
	gpio_pin_configure(gpio, LED_COL6_GPIO_PIN, GPIO_DIR_OUT);
	gpio_pin_configure(gpio, LED_COL7_GPIO_PIN, GPIO_DIR_OUT);
	gpio_pin_configure(gpio, LED_COL8_GPIO_PIN, GPIO_DIR_OUT);
	gpio_pin_configure(gpio, LED_COL9_GPIO_PIN, GPIO_DIR_OUT);
}

void uint32ToAscii(uint32_t hex, char* ascii)
{
	int i;

	for (i = 0; i < 8; i++)
	{
		//char c = toupper(hex[i]);
		uint8_t temp = ( hex >> ( i * 4 ) ) & 0x0f;

		if (temp >= 0 && temp <= 9)
		{
			ascii[7 - i] = temp + 0x30;

		}
		else{
			ascii[7 - i] = temp + 55;
		}
	}
	ascii[8] = '\0';

}


static void bt_ready(int err)
{
	uint32_t uuidFICR = NRF_FICR->DEVICEADDR[0] ;
	char deviceName[13] = "BTA-";

	uint32ToAscii(uuidFICR, (char*)&deviceName[4]);
	printk("Device Name: %s \n", deviceName);


	if (err) {
		printk("Bluetooth init failed (err %d)\n", err);
		return;
	}


	board_init();

	/* Use identity address as device UUID */
	memcpy(dev_uuid, &uuidFICR, sizeof(uint64_t));

	/* Use "BTA-xxxxxxxx" as Bluetooth device name */
	err = bt_set_name((const char*)deviceName);
	if(err){
		printk("bt_set_name (err %d)\n", err);
		return;
	}

	err = bt_mesh_init(&prov, &comp);
	if (err) {
		printk("Initializing mesh failed (err %d)\n", err);
		return;
	}

	bt_mesh_prov_enable(/*BT_MESH_PROV_ADV |*/ BT_MESH_PROV_GATT);

}

void main(void)
{
	int err;

	printk("Bluetooth Asia 2019 - Bluetooth mesh hands-on...\n");

	/* Initialize the Bluetooth Subsystem */
	err = bt_enable(bt_ready);
	if (err) {
		printk("Bluetooth init failed (err %d)\n", err);
	}
}
